"""A python implementation of an mzML file reader"""

import zlib
import base64
import struct
import warnings
from copy import copy
from xml.sax import parse
from xml.sax.handler import ContentHandler
import pynumpress
from rawfilereader import Analyzer, ScanMode, ScanType, Polarity, accessions
from rawfilereader.main_dataclasses import Scan, ScanFilter, RawFile

COMPRESSION_TYPES = {
    "no compression": lambda x: x,
    "zlib": zlib.decompress,
    "numpress linear": pynumpress.decode_linear,
    "numpress positive int": pynumpress.decode_pic,
    "numpress short logged float": pynumpress.decode_slof,
}


def parse_mzml(filename_or_stream):
    """Returns a Rawfile given a file string or stream
    :param filename_or_stream:string (filename) or file object
    :return: Rawfile
    """
    parser = MzmlParser()
    parse(filename_or_stream, parser)  # nosec
    return parser.rawfile


def parse_filter_string(filter_string, scan_filter=ScanFilter()):
    """
    Fills analyzer, scan type, scan mode, and dependent scan info from a filter string.
    :param filter_string: string, filter string saved from thermo raw files
    :param scan_filter: ScanFilter, a pre-existing ScanFilter to populate, defaults to
    new ScanFilter
    :return: ScanFilter, copy of the provided ScanFilter with additional info filled
    """
    filter_string = filter_string.lower()
    scan_filter = copy(scan_filter)
    for a in Analyzer:
        if a.name.lower() in filter_string:
            scan_filter.analyzer = a
            break
    for scan_type in ScanType:
        if scan_type.name.lower() in filter_string:
            scan_filter.scan_type = scan_type
    scan_filter.dependent_scans = " d " in filter_string

    if " p " in filter_string:
        scan_filter.scan_mode = ScanMode.PROFILE
    elif " c " in filter_string:
        scan_filter.scan_mode = ScanMode.CENTROID

    return scan_filter


class MzmlParser(ContentHandler):
    """mzML parser class to be used with xml.sax"""

    def __init__(self):
        super().__init__()
        self.stack = []
        self.current_function = None
        self.rawfile = RawFile()
        self.rawfile.low_mz = -1

        self.warn = False  # scan_mode warning should or shouldn't be raised

        self.referenceable_groups = {}
        self.instrument_config = {}

        self.start_method_map = {
            "mzML": self.set_version,
            "sourceFile": self.set_file_name,
            "cvParam": self.handle_cv_param,
            "spectrumList": self.set_spectrum_count,
            "spectrum": self.create_spectrum,
        }

        self.end_method_map = {
            "referenceableParamGroup": self.handle_referenceable_group,
            "instrumentConfiguration": self.handle_instrument_config,
            "spectrum": self.finalize_spectrum,
            "mzML": self.finalize_rawfile,
            "binary": self.handle_binary_data,
            "binaryDataArray": self.finalize_binary_data_array,
        }

    def startElement(self, name, attrs):
        # callbacks, if applicable
        if self.current_function:
            self.current_function(name, dict(attrs))  # pylint: disable=not-callable
        elif name in self.start_method_map:
            method = self.start_method_map[name]
            method(name, dict(attrs))

        if name in ("indexedmzML", "mzML"):
            return
        element = {"name": name, "attrs": dict(attrs), "value": "", "children": []}
        if len(self.stack) > 0:
            self.stack[-1]["children"].append(element)
        self.stack.append(element)

    def endElement(self, name):
        if name in self.end_method_map:
            self.end_method_map[name]()

        if name in ("indexedmzML", "mzML"):
            return
        self.stack.pop()

    def characters(self, content):
        if content.strip() != "":
            self.stack[-1]["value"] += repr(content)

    def handle_cv_param(self, _, attrs):
        """Callback to handle generic cvParam elements"""
        acc = attrs["accession"]
        if acc in accessions.INSTRUMENTS:
            self.rawfile.instrument_model = attrs["name"]

    def set_version(self, _, attrs):
        """Callback for mzML element. Sets file version."""
        self.rawfile.version = attrs["version"]

    def set_file_name(self, _, attrs):
        """Callback for sourceFile. Sets file name."""
        if len(self.stack[-1]["children"]) == 0:  # first file only
            self.rawfile.file_name = attrs["name"]

    def handle_referenceable_group(self):
        """Callback for referenceableParamGroup. Stores group contents in parser."""
        group = self.stack[-1]
        self.referenceable_groups[group["attrs"]["id"]] = group["children"]

    def handle_instrument_config(self):
        """Callback for instrumentConfiguration. Stores group contents in parser."""
        group = self.stack[-1]
        self.instrument_config[group["attrs"]["id"]] = group["children"]

    def set_spectrum_count(self, _, attrs):
        """Callback for spectrumList. Sets num_scans."""
        self.rawfile.num_scans = int(attrs["count"])

    def create_spectrum(self, _, attrs):
        """Callback for spectrum element. Creates new Scan."""
        # create and store a temporary ScanFilter within the Scan
        self.rawfile.scans.append(Scan(index=int(attrs["index"]), filter=ScanFilter()))
        self.current_function = self.handle_scan_cv_params

    def handle_scan_cv_params(self, name, attrs):  # pylint: disable=too-many-branches
        """Callback for elements within a spectrum. Sets various Scan fields."""
        if name == "binaryDataArray":
            self.current_function = self.handle_binary_data_array
        if name != "cvParam":
            return
        acc = attrs["accession"]
        unit = None if "unitAccession" not in attrs else attrs["unitAccession"]
        current_scan = self.rawfile.scans[-1]
        if acc == accessions.POSITIVE_SCAN:
            current_scan.filter.polarity = Polarity.POSITIVE
        elif acc == accessions.NEGATIVE_SCAN:
            current_scan.filter.polarity = Polarity.NEGATIVE
        elif acc == accessions.MS_LEVEL:
            current_scan.filter.ms_level = int(attrs["value"])
        elif acc == accessions.PROFILE:
            current_scan.filter.scan_mode = ScanMode.PROFILE
        elif acc == accessions.CENTROID:
            current_scan.filter.scan_mode = ScanMode.CENTROID
        elif acc == accessions.TIME:
            current_scan.time = float(attrs["value"])
        elif acc == accessions.FILTER_STRING:
            current_scan.filter.filter_string = attrs["value"]
        elif acc == accessions.LOW_MZ and unit == accessions.UNIT_MZ:
            current_scan.filter.low_mass = float(attrs["value"])
            if self.rawfile.low_mz < 0 or float(attrs["value"]) < self.rawfile.low_mz:
                self.rawfile.low_mz = float(attrs["value"])
        elif acc == accessions.HIGH_MZ and unit == accessions.UNIT_MZ:
            current_scan.filter.high_mass = float(attrs["value"])
            if float(attrs["value"]) > self.rawfile.high_mz:
                self.rawfile.high_mz = float(attrs["value"])
        # consider "selected ion m/z" MS:1000744 if isolation window is not reliable
        elif acc == accessions.ISOLATION_WINDOW and unit == accessions.UNIT_MZ:
            current_scan.filter.precursor_mz.append(float(attrs["value"]))
        elif acc == accessions.COLLISION_ENERGY:
            current_scan.filter.energy.append(float(attrs["value"]))

    def finalize_spectrum(self):
        """
        Callback for spectrum element. Finalizes ScanFilter and sets default values if
        needed.
        """
        self.current_function = None
        current_scan = self.rawfile.scans[-1]

        if len(current_scan.filter.precursor_mz) == 0:
            current_scan.filter.precursor_mz = [
                0 for _ in range(current_scan.filter.ms_level)
            ]
        if len(current_scan.filter.energy) == 0:
            current_scan.filter.energy = [
                0 for _ in range(current_scan.filter.ms_level)
            ]
        if len(current_scan.intensities) == 0:  # no profile data
            current_scan.intensities = current_scan.cent_intensities
            current_scan.mzs = current_scan.cent_mzs

        new_filter = parse_filter_string(
            current_scan.filter.filter_string, current_scan.filter
        )
        if (
            new_filter.scan_mode == ScanMode.PROFILE
            and current_scan.filter.scan_mode == ScanMode.CENTROID
        ):
            self.warn = True
        current_scan.filter = new_filter

        if current_scan.filter not in self.rawfile.scan_filters:
            self.rawfile.scan_filters.append(current_scan.filter)
        current_scan.filter = self.rawfile.scan_filters.index(current_scan.filter)

    def handle_binary_data_array(self, name, attrs):
        """
        Callback for binaryDataArray within a spectrum. Sets units, compression, and
        binary data.
        """
        if name != "cvParam":
            return

        acc = attrs["accession"]
        if acc in accessions.BINARY_UNITS:
            self.stack[-1]["unit"] = accessions.BINARY_UNITS[acc]
        elif acc in accessions.BINARY_COMPRESSION:
            self.stack[-1]["compression"] = accessions.BINARY_COMPRESSION[acc]
        elif acc == accessions.INTENSITY_ARRAY:
            self.stack[-1]["array type"] = "intensity"
        elif acc == accessions.MZ_ARRAY:
            self.stack[-1]["array type"] = "mz array"

    def handle_binary_data(self):
        """Callback for binary element"""
        if self.stack[-2]["name"] != "binaryDataArray":
            return
        # copy binary data to parent in stack
        self.stack[-2]["binary"] = self.stack[-1]["value"]

    def finalize_binary_data_array(self):
        """
        Callback for binaryDataArray. Decompresses and decodes m/z and intensity data.
        """
        # ignore chromatograms
        if len(self.stack) < 3 or self.stack[-3]["name"] != "spectrum":
            return
        self.current_function = self.handle_scan_cv_params  # back to spectrum
        data_array = self.stack[-1]
        if "array type" not in data_array:
            # is not intensity or mz, can be skipped
            return
        if "unit" not in data_array:
            raise RuntimeError("Binary unit missing or not supported.")
        if "compression" not in data_array:
            raise RuntimeError("Binary compression type missing or not supported.")

        binary = base64.b64decode(self.stack[-1]["binary"])
        for compression in data_array["compression"]:
            func = COMPRESSION_TYPES[compression]
            binary = func(binary)

        unit = data_array["unit"]
        if "numpress" in data_array["compression"][-1]:
            decoded = binary
        elif unit:
            format_string = f"<{int(len(binary) / unit[1])}{unit[0]}"
            decoded = struct.unpack(format_string, binary)
        else:
            decoded = binary.decode("ascii")

        if (
            data_array["array type"] == "intensity"
            and self.rawfile.scans[-1].filter.scan_mode == ScanMode.PROFILE
        ):
            self.rawfile.scans[-1].intensities = list(decoded)
        elif data_array["array type"] == "intensity":
            self.rawfile.scans[-1].cent_intensities = list(decoded)
        elif (
            data_array["array type"] == "mz array"
            and self.rawfile.scans[-1].filter.scan_mode == ScanMode.PROFILE
        ):
            self.rawfile.scans[-1].mzs = list(decoded)
        else:
            self.rawfile.scans[-1].cent_mzs = list(decoded)

    def finalize_rawfile(self):
        """Callback for close of mzML element"""
        self.rawfile.file_format = "mzML"
        self.rawfile.first_scan_number = self.rawfile.scans[0].index
        self.rawfile.last_scan_number = self.rawfile.scans[-1].index
        self.rawfile.start_time = self.rawfile.scans[0].time
        self.rawfile.end_time = self.rawfile.scans[-1].time

        if self.warn:
            warnings.warn(
                "One or more scans are labeled as being acquired in profile mode but "
                + "only contain centroid data."
            )



# pylint: disable=too-many-instance-attributes,missing-class-docstring
"""
Package to open and analyze raw ms files
"""
import ctypes
from ctypes import (
    CDLL,
    c_void_p,
    c_char_p,
    c_float,
    c_double,
    c_uint8,
    c_int16,
    c_int32,
    c_bool,
    c_uint32,
    c_uint64,
    Structure,
    POINTER,
    cast,
)
import os
import platform
from enum import Enum
import numpy as np

__version__ = "0.0.3"


dir_path = os.path.dirname(os.path.realpath(__file__))
if platform.system() == "Windows":
    rawfilereader = CDLL(os.path.join(dir_path, "rawfilereader.dll"))
else:
    rawfilereader = CDLL(os.path.join(dir_path, "rawfilereader.so"))


class Empty:
    pass


class Analyzer(Enum):
    ITMS = 0
    TQMS = 1
    SQMS = 2
    TQFMS = 3
    FTMS = 4
    SECTOR = 5
    UNDEFINED = 6


class ScanMode(Enum):
    CENTROID = 0
    PROFILE = 1
    UNDEFINED = 2


class ScanType(Enum):
    FULL = 0
    ZOOM = 1
    SIM = 2
    SRM = 3
    CRM = 4
    UNDEFINED = 5
    Q1 = 6
    Q3 = 7


class Polarity(Enum):
    NEGATIVE = 0
    POSITIVE = 1
    UNDEFINED = 2


class CXic(Structure):  # pylint: disable=too-few-public-methods
    """
    C Struct for the XIC
    """

    _fields_ = [
        ("rt", c_void_p),
        ("intensity", c_void_p),
        ("length", c_int32),
        ("mzs", c_void_p),
    ]


class CScanFilter(Structure):  # pylint: disable=too-few-public-methods
    """
    C Struct for the ScanFilter
    """

    _fields_ = [
        ("polarity", c_uint8),
        ("analyzer", c_uint8),
        ("ms_level", c_uint8),
        ("n_precursors", c_uint32),
        ("precursor_mz", c_void_p),
        ("energy", c_void_p),
        ("low_mass", c_double),
        ("high_mass", c_double),
        ("scan_mode", c_uint8),
        ("scan_type", c_uint8),
        # ("dependent_scans", c_uint8),
    ]


class CRawFile(Structure):  # pylint: disable=too-few-public-methods
    """
    C Struct for the RawFile
    """

    _fields_ = [
        ("instrument_model", c_char_p),
        ("filename", c_char_p),
        ("n_controllers", c_uint32),
        ("scans", c_void_p),
        ("data", c_void_p),  # pointer to full data, instead of raw bytes
        ("num_scans", c_uint32),
        ("first_scan_number", c_uint32),
        ("last_scan_number", c_uint32),
        ("start_time", c_double),
        ("end_time", c_double),
        ("scan_filters", c_void_p),
        ("n_filters", c_uint32),
        ("timestamp", c_uint64),
        ("chromatograms", c_void_p),
        ("num_chroms", c_uint32),
        # ("scan_trailer_addr", c_uint64),
        # ("low_mz", c_double),
        # ("high_mz", c_double),
        # ("scan_index_addr", c_uint64),
        # ("scan_params_addr", c_uint64),
        # ("version", c_uint32),
        # ("data_addr", c_uint64),
    ]


class CScan(Structure):  # pylint: disable=too-few-public-methods
    """
    C Struct for Scan
    """

    _fields_ = [
        ("prIntensities", c_void_p),
        ("prMzs", c_void_p),
        ("centIntensities", c_void_p),
        ("centMzs", c_void_p),
        ("prTotal", c_uint32),
        ("centTotal", c_uint32),
        ("time", c_double),
        ("index", c_uint32),
        ("filter", c_uint32),
    ]


class CChrom(Structure):  # pylint: disable=too-few-public-methods
    """
    C Struct for Chromatogram
    """

    _fields_ = [
        ("id", c_void_p),
        ("intensities", c_void_p),
        ("rts", c_void_p),
        ("length", c_uint32),
        ("precursorMz", c_float),
        ("productMz", c_float),
    ]


# Define the arg and return types of the function
rawfilereader.Open.argtypes = [c_char_p, c_int32, c_int32, c_char_p, c_uint64]
rawfilereader.Open.restype = c_void_p  # RawFile, just store as a pointer

rawfilereader.FromBytes.argtypes = [c_char_p, c_int32, c_int32, c_char_p]
rawfilereader.FromBytes.restype = c_void_p  # RawFile, just store as a pointer

rawfilereader.FreeRawFile.argtypes = [c_void_p]
rawfilereader.FreeRawFile.restype = c_int16

rawfilereader.Pull_xic.argtypes = [
    c_void_p,
    c_float,
    c_float,
    c_float,
    c_float,
    c_int32,
    c_bool,
    c_bool,
]
rawfilereader.Pull_xic.restype = CXic

rawfilereader.FreeP.argtypes = [c_void_p]

rawfilereader.GetErrorString.restype = c_char_p


# Define our usable structures
class Xic:  # pylint: disable=too-few-public-methods
    """
    An Extracted Ion Chromatogram. Mainly, it stores numpy arrays for
    retention times and intensities
    """

    def __init__(self, c_xic=None):
        """Initialize from a C_Xic struct"""
        if c_xic is None:
            self.length = 0
            self.rt = np.array([])
            self.intensity = np.array([])
            self.mz = np.array([])
            return
        if isinstance(c_xic, dict):
            self.rt = c_xic["rt"]
            self.intensity = c_xic["intensity"]
            self.length = c_xic["length"]
            self.mz = np.array([c_xic["mz"]] * c_xic["length"])
            return

        rt_p = c_xic.rt
        intensity_p = c_xic.intensity
        mz_p = c_xic.mzs
        self.length = c_xic.length
        self.rt = np.copy(
            np.frombuffer((c_float * c_xic.length).from_address(rt_p), np.float32)
        )
        self.intensity = np.copy(
            np.frombuffer(
                (c_float * c_xic.length).from_address(intensity_p), np.float32
            )
        )
        if mz_p:
            self.mz = np.copy(
                np.frombuffer((c_float * c_xic.length).from_address(mz_p), np.float32)
            )
        else:
            self.mz = np.zeros(self.length, np.float32)

        rawfilereader.FreeP(rt_p)
        rawfilereader.FreeP(intensity_p)
        rawfilereader.FreeP(mz_p)


class ScanFilter:  # pylint: disable=too-few-public-methods
    """
    A scan filter, an assortment of properties that a scan can have
    """

    def __init__(self, scanf):
        self.polarity = Polarity(scanf["polarity"])
        self.scan_mode = ScanMode(scanf["scan_mode"])
        # self.dependent_scans = scanf["dependent_scans"]
        self.ms_level = scanf["ms_level"]
        self.scan_type = ScanType(scanf["scan_type"])
        self.analyzer = Analyzer(scanf["analyzer"])
        self.n_precursors = scanf["n_precursors"]
        self.low_mass = scanf["low_mass"]
        self.high_mass = scanf["high_mass"]

        # with such short lists, this seems to be faster than np.frombuffer
        self.precursor_mz = list(
            (c_double * self.n_precursors).from_address(int(scanf["precursor_mz"]))
        )
        self.energy = list(
            (c_double * self.n_precursors).from_address(int(scanf["energy"]))
        )

    def __str__(self):
        if self.ms_level == 1:
            return f"MS1; Range:{self.low_mass}-{self.high_mass}"
        return (
            f"MS{self.ms_level}; Precursors:{self.precursor_mz}; Energy:{self.energy};"
            f" Range:{self.low_mass}-{self.high_mass}"
        )

    def __repr__(self):
        return self.__str__()


class Scan:  # pylint: disable=too-few-public-methods
    """
    An individual scan
    """

    def __init__(self, c_scan, profile=False, centroid=True):
        self.profile = profile
        self.centroid = centroid
        self.time = c_scan["time"]
        self.index = c_scan["index"]
        self.cent_total = int(c_scan["centTotal"])
        self.pr_total = int(c_scan["prTotal"])
        self.scan_filter = c_scan["filter"]

        if self.centroid:
            self.cent_mzs = np.frombuffer(
                (c_float * c_scan["centTotal"]).from_address(int(c_scan["centMzs"])),
                c_float,
            )
            self.cent_intensities = np.frombuffer(
                (c_float * c_scan["centTotal"]).from_address(
                    int(c_scan["centIntensities"])
                ),
                c_float,
            )
        if self.profile:
            self.profile_mzs = np.frombuffer(
                (c_float * c_scan["prTotal"]).from_address(int(c_scan["prMzs"])),
                c_float,
            )
            self.profile_intensities = np.frombuffer(
                (c_float * c_scan["prTotal"]).from_address(
                    int(c_scan["prIntensities"])
                ),
                c_float,
            )


class Chromatogram:  # pylint: disable=too-few-public-methods
    """
    An individual scan
    """

    def __init__(self, c_chrom):
        self.id = ctypes.string_at(int(c_chrom["id"]))
        self.length = int(c_chrom["length"])
        self.precursor_mz = c_chrom["precursorMz"]
        self.product_mz = c_chrom["productMz"]

        self.rts = np.frombuffer(
            (c_float * c_chrom["length"]).from_address(int(c_chrom["rts"])),
            c_float,
        )
        self.intensities = np.frombuffer(
            (c_float * c_chrom["length"]).from_address(int(c_chrom["intensities"])),
            c_float,
        )


class RawFile:
    """
    The Open rawfile. Constructor accepts a filename string or file contents as bytes.
    """

    def __init__(self, file, profile=False, centroid=True, file_format=None):
        if isinstance(file, str):
            file_size = os.path.getsize(file)
            if file_format is None:
                if file.lower().endswith(".raw"):
                    file_format = "rawfile"
                else:
                    file_format = "mzml"
            format_char = file_format.encode("utf-8")
            filename_char = file.encode("utf-8")
            ptr = rawfilereader.Open(
                filename_char, profile, centroid, format_char, file_size
            )
        elif isinstance(file, bytes):
            if file_format is None:
                file_format = "rawfile"
            format_char = file_format.encode("utf-8")
            ptr = rawfilereader.FromBytes(file, profile, centroid, format_char)
        else:
            raise TypeError(
                "RawFile can be constructed from a string (filename) or bytes "
                f"(file contents). You provided {type(file)}."
            )
        if not ptr:  # null pointer if reading the raw file failed
            raise RuntimeError(rawfilereader.GetErrorString().decode("UTF-8"))
        rf = cast(ptr, POINTER(CRawFile)).contents  # dereference pointer

        self._data_p = ptr
        try:
            self.instrument_model = rf.instrument_model.decode("UTF-8")
        except:
            self.instrument_model = "unknown"
        try:
            self.filename = rf.filename.decode("UTF-8")
        except UnicodeDecodeError:
            self.filename = ""
        self.profile = profile
        self.centroid = centroid
        self.data = rf.data  # pointer
        self.num_scans = rf.num_scans
        self.num_chroms = rf.num_chroms

        # self.scan_trailer_addr = rf.scan_trailer_addr
        self.first_scan_number = rf.first_scan_number
        self.last_scan_number = rf.last_scan_number
        # self.low_mz = rf.low_mz
        # self.high_mz = rf.high_mz
        self.start_time = rf.start_time
        self.end_time = rf.end_time
        # self.scan_index_addr = rf.scan_index_addr
        # self.scan_params_addr = rf.scan_params_addr
        # self.version = rf.version
        self.n_controllers = rf.n_controllers
        # self.data_addr = rf.data_addr
        self.n_filters = rf.n_filters
        self.timestamp = rf.timestamp
        self.scan_filters = self._get_scan_filters(
            rf.scan_filters
        )  # number of filters,
        if self.num_scans > 0:
            scans = np.frombuffer(
                (CScan * self.num_scans).from_address(rf.scans), CScan
            )
            self.scans = [Scan(s, profile, centroid) for s in scans]
        else:
            self.scans = []

        if self.num_chroms > 0:
            chromatograms = np.frombuffer(
                (CChrom * self.num_chroms).from_address(rf.chromatograms),
                CChrom,
            )
            self.chromatograms = [Chromatogram(c) for c in chromatograms]

        else:
            self.chromatograms = []

    def __del__(self):
        try:
            rawfilereader.FreeRawFile(self._data_p)
        except AttributeError:
            # raw file failed to initialize, just move on
            pass

    def _get_scan_filters(self, ptr):
        if not ptr:
            return []
        filters = np.copy(
            np.frombuffer((CScanFilter * self.n_filters).from_address(ptr), CScanFilter)
        )
        return [ScanFilter(f) for f in filters]

    def chrom_xic(self, mz, precursor, rt1, rt2):
        """
        Given a product mz and precursor mz and rt start and stop,
        returns a chromtogram. Look for +/- 0.5 mz for ranges
        """
        if self.num_chroms == 0:
            return Xic(None)
        for chrom in self.chromatograms:
            if (
                abs(chrom.precursor_mz - precursor) < 0.5
                and abs(chrom.product_mz - mz) < 0.5
            ):
                start = np.searchsorted(chrom.rts, rt1, side="left")
                stop = np.searchsorted(chrom.rts, rt2, side="right")
                if start == len(chrom.rts) or stop == 0:
                    return Xic(None)
                return Xic(
                    {
                        "rt": chrom.rts[start:stop],
                        "intensity": chrom.intensities[start:stop],
                        "length": stop - start,
                        "mz": chrom.product_mz,
                    }
                )
        return Xic(None)

    def xic(
        self,
        mz,
        rt1,
        rt2,
        ppm=20,
        scan_filter=None,
        centroid=True,
        pull_mzs=False,
        precursor=None,
    ):
        """
        Pulls an XIC
        """

        if self.num_scans == 0:
            return self.chrom_xic(mz, precursor, rt1, rt2)
        if scan_filter is None:
            scan_filter = -1

        if (centroid and not self.centroid) or (not centroid and not self.profile):
            empty = Empty()
            empty.intensity = np.array([])
            empty.rt = np.array([])
            empty.length = 0
            return empty
        return Xic(
            rawfilereader.Pull_xic(
                self._data_p, rt1, rt2, mz, ppm, scan_filter, centroid, pull_mzs
            )
        )

    def ms2(self, mz, rt1, rt2, ppm=5, centroid=True):
        """
        Pulls MS2 fragments for a given precursor MZ and RT range
        Returns fragment data and info about the scans and scan filters used
        """
        if not centroid:
            return None

        mz1 = mz - (mz / 1000000) * ppm
        mz2 = mz + (mz / 1000000) * ppm

        filters = []
        for i, scan_filter in enumerate(self.scan_filters):
            if (
                len(scan_filter.precursor_mz) > 0
                and mz1 <= scan_filter.precursor_mz[0] <= mz2
            ):
                filters.append(i)

        scan_list = []
        polarity = set()
        for i, scan in enumerate(self.scans):
            if scan.scan_filter in filters and rt1 <= scan.time <= rt2:
                scan_list.append(scan)
                polarity.add(self.scan_filters[scan.scan_filter].polarity)

        scan_info = {
            # filename without full path, if possible
            "rawfile": [self.filename.split("\\")[-1]],
            "scanList": [scan.index for scan in scan_list],
            "scanRts": [scan.time for scan in scan_list],
            "massRange": [0, 0],
            "precursorRange": [mz1, mz2],
            "resolution": [],
            "polarity": [p.name.lower() for p in polarity],
        }

        raw = []
        for scan in scan_list:
            for ms2_mz, intensity in zip(scan.cent_mzs, scan.cent_intensities):
                scan_filter = self.scan_filters[scan.scan_filter]
                scan_info["massRange"] = [scan_filter.low_mass, scan_filter.high_mass]
                raw.append(
                    {
                        "scanNumber": scan.index,
                        "rt": scan.time,
                        "mz": ms2_mz,
                        "intensity": intensity,
                        "chemicalFormula": "",
                        "formulaPpm": "",
                        "NCE": scan_filter.energy[0],
                        "aggregateNumber": "",
                    }
                )

        return raw, scan_info

    def export_ms2(
        self,
        mz,
        rt1,
        rt2,
        ppm=5,
        filename=None,
        rt_prec=4,
        mz_prec=5,
        int_prec=0,
        min_scans=1,
        include_raw=True,
    ):
        """
        Extracts, aggregates, and exports MS2 data in a standard MXP reporting format.
        Writes to csv if filename is provided, including extension (e.g. "results.csv").
        """
        raw, header = self.ms2(mz, rt1, rt2, ppm)
        raw, aggregated = self.aggregate(raw, ppm)

        output = [["HEADER"], ["version", 1]]
        for key, value in header.items():
            if isinstance(value, list):
                output.append([key, *value])
            else:
                output.append([key, value])

        output.append(["RAW"])
        if include_raw and len(raw) > 0:
            output.append(list(raw[0].keys()))
            for scan in raw:
                temp_output = list(scan.values())
                temp_output[1] = f"{temp_output[1]:.{rt_prec}f}"
                temp_output[2] = f"{temp_output[2]:.{mz_prec}f}"
                temp_output[3] = f"{temp_output[3]:.{int_prec}f}"
                output.append(temp_output)

        output.append(["AGGREGATED", "sum_v2"])
        if len(aggregated) > 0:
            output.append(list(aggregated[0].keys()))
            for group in aggregated:
                temp_output = list(group.values())
                if temp_output[6] < min_scans:
                    continue
                temp_output[1] = f"{temp_output[1]:.{mz_prec}f}"
                temp_output[2] = f"{temp_output[2]:.{int_prec}f}"
                output.append(temp_output)

        if filename:
            with open(filename, "w", encoding="ISO-8859-1") as file:
                for line in output:
                    file.write(",".join([str(item) for item in line]) + "\n")

        return output

    def aggregate(self, raw, ppm):
        """
        Aggregate MSn fragments with similar masses and equal NCEs.
        :param raw: list of dicts; MSn fragment data, including keys: "mz", "intensity",
                    "chemicalFormula", "formulaPpm", "NCE" at minimum
        :param ppm: float, permitted delta ppm for grouping MZs
        :return:
            raw: list of dicts, raw input updated with aggregateNumber
            aggregated: list of dicts, aggregated fragments
        """
        aggregated = []
        if not raw:
            return raw, aggregated
        raw_copy = raw.copy()  # Clearer than raw[:]
        raw_copy.sort(reverse=True, key=lambda e: e["intensity"])
        active = np.ones(len(raw_copy), dtype=bool)
        group_idx = 0
        data = np.array([(e["mz"], e["intensity"], e["NCE"]) for e in raw_copy])

        for i, raw_feature in enumerate(raw_copy):
            if not active[i]:
                continue

            center_mz = data[i, 0]
            ppm_factor = (center_mz / 1_000_000) * ppm
            min_mz, max_mz = center_mz - ppm_factor, center_mz + ppm_factor
            energy = data[i, 2]
            matched = (
                active
                & (data[:, 2] == energy)
                & (data[:, 0] >= min_mz)
                & (data[:, 0] <= max_mz)
            )
            if not matched.any():
                continue

            average_mz = np.average(data[matched, 0], weights=data[matched, 1])
            aggregated.append(
                {
                    "aggregateNumber": group_idx,
                    "mz": average_mz,
                    "intensity": data[matched, 1].sum(),
                    "chemicalFormula": raw_feature["chemicalFormula"],
                    "formulaPpm": raw_feature["formulaPpm"],
                    "NCE": energy,
                    "scanCount": matched.sum(),
                }
            )
            matched_indices = np.where(matched)[0]
            for j in matched_indices:
                raw_copy[j]["aggregateNumber"] = group_idx
            group_idx += 1
            active[matched_indices] = False
        return raw, aggregated


# testing functions
if __name__ == "__main__":
    from datetime import datetime, timedelta

    rf = RawFile(r"C:\Users\danie\Downloads\prepare.raw", profile=True)
    print(rf.timestamp)
    print(datetime(1601, 1, 1) + timedelta(microseconds=rf.timestamp / 10))
    print(datetime(1601, 1, 1) + timedelta(microseconds=rf.timestamp // 10))



# pylint: disable=redefined-outer-name
"""
Tests for MSEclipse
"""
from pathlib import Path
import pickle
import numpy as np
import pandas as pd
import bmxp.eclipse as ms
import networkx as nx
import pytest


@pytest.fixture()
def filepath_1():
    """
    First file path for test
    """
    return Path(__file__).parent / "test1.csv"


@pytest.fixture()
def filepath_2():
    """
    Second file path for test
    """
    return Path(__file__).parent / "test2.csv"


@pytest.fixture()
def filepath_3():
    """
    Third file path for test
    """
    return Path(__file__).parent / "test3.csv"


@pytest.fixture()
def dataframe_1(filepath_1):
    """
    First dataframe for test
    """
    return pd.read_csv(filepath_1)


@pytest.fixture()
def dataframe_2(filepath_2):
    """
    Second dataframe for test
    """
    return pd.read_csv(filepath_2)


@pytest.fixture()
def dataframe_3(filepath_3):
    """
    Third dataframe for test
    """
    return pd.read_csv(filepath_3)


@pytest.fixture()
def pickled_ms():
    """
    Load pickled file
    """
    return pickle.load(
        open(
            Path(__file__).parent / "mseclipse.pickle", "rb"
        )  # pylint: disable=consider-using-with
    )


@pytest.fixture()
def integration_data():
    """
    Load pickled file
    """
    pickles = []
    datasets = []
    for f in [
        "full_cliques.pickle",
        "ranked_cliques.pickle",
        "all_cliques.pickle",
        "1_non_clique.pickle",
    ]:
        df = pickle.load(
            open(Path(__file__).parent / f, "rb")  # pylint: disable=consider-using-with
        )
        df = df.sort_values(by=["DS1", "DS2", "DS3", "DS4"]).reset_index(drop=True)
        pickles.append(df)
    for f in ["DS1.csv", "DS2.csv", "DS3.csv", "DS4.csv"]:
        datasets.append(pd.read_csv(Path(__file__).parent / f))
    return [pickles, datasets]


def test_initialize_add(filepath_1, filepath_2, dataframe_1, dataframe_2):
    """
    Test MSAligner initializes or throws errors
    """

    # example test: show that importing and utilizing the class is successful
    ms.MSAligner(filepath_1, filepath_2, names=["hp1", "hp2"])

    # Check that dataframes can be read
    ms.MSAligner(dataframe_1, dataframe_2, names=["hp1", "hp2"])

    # datasets require names
    with pytest.raises(ValueError) as e:
        ms.MSAligner(dataframe_1, dataframe_2)
    assert "must be provided" in str(e.value)

    # check mismatch in length
    with pytest.raises(ValueError) as e:
        ms.MSAligner(filepath_1, filepath_2, names=["hp1", "hp2", "hp3"])
    assert "The number of names" in str(e.value)

    # check duplicate names
    with pytest.raises(ValueError) as e:
        ms.MSAligner(filepath_1, filepath_2, names=["hp2", "hp2"])
    assert "duplicate" in str(e.value)


def test_add_dataset(filepath_1, filepath_2, dataframe_1, dataframe_2):
    """
    Test MSAligner add_dataset method
    """
    a = ms.MSAligner()

    # example test: show that importing and utilizing the class is successful
    a.add_dataset(filepath_1, "hp1")
    a.add_dataset(dataframe_2, "hp2")

    # check duplicate names
    with pytest.raises(ValueError) as e:
        a.add_dataset(filepath_2, "hp2")
    assert "duplicate" in str(e.value)

    # test dataframe can be added
    a.add_dataset(dataframe_1, "hp3")

    # dataframes require names
    with pytest.raises(ValueError) as e:
        a.add_dataset(dataframe_2)
    assert "must be provided" in str(e.value)


def test_alignment_methods(dataframe_1, dataframe_2, pickled_ms):
    """
    Test the methods associated with alignment
    """

    a = ms.MSAligner(dataframe_1, dataframe_2, names=["HP1", "HP2"])

    # Test anchors
    a.gen_anchors()
    assert a.anchors["HP1"] == pickled_ms.anchors["HP1"]
    assert a.anchors["HP2"] == pickled_ms.anchors["HP2"]

    # Test coarse matching
    a.gen_coarse()
    assert np.equal(
        a.coarse_matches["HP1"]["HP2"].values,
        pickled_ms.coarse_matches["HP1"]["HP2"].values,
    ).all()

    # Test scaler generation; RT, MZ, and intensity
    a.gen_scalers()
    hp1_scalers = a.scalers["HP1"]["HP2"]
    pickled_scalers = pickled_ms.scalers["HP1"]["HP2"]
    assert np.isclose(hp1_scalers["RT"]["y"], pickled_scalers["RT"]["y"]).all()
    assert np.isclose(hp1_scalers["RT"]["x"], pickled_scalers["RT"]["x"]).all()
    assert np.isclose(hp1_scalers["MZ"]["y"], pickled_scalers["MZ"]["y"]).all()
    assert np.isclose(hp1_scalers["MZ"]["x"], pickled_scalers["MZ"]["x"]).all()
    assert np.isclose(
        hp1_scalers["Intensity"]["x"], pickled_scalers["Intensity"]["x"]
    ).all()
    assert np.isclose(
        hp1_scalers["Intensity"]["y"], pickled_scalers["Intensity"]["y"]
    ).all()

    # Test scaling functions
    a.gen_scaled_values()
    scaled_values = a.scaled_values["HP1"]["HP2"]
    pickled_values = pickled_ms.scaled_values["HP1"]["HP2"]
    assert np.isclose(scaled_values["RT"], pickled_values["RT"]).all()
    assert np.isclose(scaled_values["MZ"], pickled_values["MZ"]).all()
    assert np.isclose(scaled_values["Intensity"], pickled_values["Intensity"]).all()

    # Test standard deviation calculations
    a.gen_stds()
    assert np.isclose(a.stds["HP1"]["HP2"]["RT"], pickled_ms.stds["HP1"]["HP2"]["RT"])
    assert np.isclose(a.stds["HP1"]["HP2"]["MZ"], pickled_ms.stds["HP1"]["HP2"]["MZ"])
    assert np.isclose(
        a.stds["HP1"]["HP2"]["Intensity"], pickled_ms.stds["HP1"]["HP2"]["Intensity"]
    )

    # Test matches
    a.gen_matches()
    assert a.matches["HP1"]["HP2"].equals(pickled_ms.matches["HP1"]["HP2"])

    # Check that align passes
    a = ms.MSAligner(dataframe_1, dataframe_2, names=["HP1", "HP2"])
    a.align()
    a.results()

    # Check that custom scalers are not overwritten
    rt = ms.calc_scalers([1, 2, 3], [1.1, 2.2, 3.3])
    mz = ms.calc_scalers([100, 200, 300], [100.0005, 200.0008, 300.0009])
    intensity = ms.calc_scalers([2, 3, 4], [3, 4, 5])

    a = ms.MSAligner(dataframe_1, dataframe_2, names=["HP1", "HP2"])
    a.set_scalers(
        {
            "HP1": {
                "HP2": {
                    "RT": zip(*rt),
                    "MZ": zip(*mz),
                    "Intensity": zip(*intensity),
                }
            }
        },
        rec=False,
    )
    a.align()
    a_scalers = a.scalers["HP1"]["HP2"]
    assert np.equal(a_scalers["RT"]["x"], rt[0]).all()
    assert np.equal(a_scalers["RT"]["y"], rt[1]).all()
    assert np.equal(a_scalers["MZ"]["x"], mz[0]).all()
    assert np.equal(a_scalers["MZ"]["y"], mz[1]).all()
    assert np.equal(a_scalers["Intensity"]["x"], intensity[0]).all()
    assert np.equal(a_scalers["Intensity"]["y"], intensity[1]).all()

    # check reciprocity for scaler generation works
    a = ms.MSAligner(dataframe_1, dataframe_2, names=["HP1", "HP2"])
    a.set_scalers(
        {
            "HP1": {
                "HP2": {
                    "RT": zip(*rt),
                    "MZ": zip(*mz),
                    "Intensity": zip(*intensity),
                }
            }
        },
        rec=True,
    )
    a.align()
    a_scalers = a.scalers["HP2"]["HP1"]
    assert np.equal(a_scalers["RT"]["x"], rt[0] + rt[1]).all()
    assert np.equal(a_scalers["RT"]["y"], -rt[1]).all()
    new_mzs = mz[0] + mz[0] * mz[1] / 1000000
    assert np.equal(a_scalers["MZ"]["x"], new_mzs).all()
    assert np.equal(a_scalers["MZ"]["y"], -mz[1] * mz[0] / new_mzs).all()
    assert np.equal(a_scalers["Intensity"]["x"], intensity[0] + intensity[1]).all()
    assert np.equal(a_scalers["Intensity"]["y"], -intensity[1]).all()


def test_calc_score():
    """
    Test score calculation
    """
    score = ms.calc_score(
        {"RT": 1, "MZ": 200, "Intensity": 100},
        {"RT": 2, "MZ": 200.0002, "Intensity": 1000},
        {"RT": "linear", "MZ": "ppm", "Intensity": "log10"},
        {"RT": 1, "MZ": 1, "Intensity": 1},
        {"RT": 1, "MZ": 1, "Intensity": 1},
    )
    assert np.isclose(score, 3)


@pytest.mark.filterwarnings("ignore:invalid value")
def test_calc_scalers():
    """
    Supplemental tests to calculating scalers
    """

    # Will return either nans, the original y_vals, or 1s.
    # Test it does not make an infinite loop
    x_vals = np.array([0, 0, 1])
    y_vals = np.array([1, 1, 2])
    results = ms.calc_scalers(x_vals, y_vals, smoothing="lowess", mode="linear")[1]
    all_nans = np.isnan(results).all()
    is_original = np.isclose(np.array([1, 1, 1]), results).all()
    assert all_nans or is_original


def test_adding_intensity_column(dataframe_3):
    """
    Tests to see if intensity is autocalculated when missing
    """
    # Create Dataframes without Intensity
    df_3_no_intensity = dataframe_3.drop(columns="Intensity")

    a = ms.MSAligner()
    a.add_dataset(df_3_no_intensity, "df3")
    # Test that _calc_intensity functions on its own
    assert np.isclose(a.datasets["df3"]["Intensity"], dataframe_3["Intensity"]).all()


def test_dataset_selection(dataframe_1, dataframe_2, dataframe_3):
    """
    Tests the wrappers which control the inner and outer dataframe for loops
    """
    # test all works
    names = ["HP1", "HP2", "HP3"]
    a = ms.MSAligner(dataframe_1, dataframe_2, dataframe_3, names=names)
    a.align()
    assert list(a.matches) == ["HP1", "HP2", "HP3"]
    assert list(a.matches["HP1"]) == ["HP2", "HP3"]
    assert list(a.matches["HP2"]) == ["HP1", "HP3"]
    assert list(a.matches["HP3"]) == ["HP1", "HP2"]

    # test list to all
    a = ms.MSAligner(dataframe_1, dataframe_2, dataframe_3, names=names)
    a.align(["HP1", "HP2", "HP3"])
    assert list(a.matches) == ["HP1", "HP2", "HP3"]
    assert list(a.matches["HP1"]) == ["HP2", "HP3"]
    assert list(a.matches["HP2"]) == ["HP1", "HP3"]
    assert list(a.matches["HP3"]) == ["HP1", "HP2"]

    # test one way to all
    a = ms.MSAligner(dataframe_1, dataframe_2, dataframe_3, names=names)
    a.align("HP1")
    assert list(a.matches) == ["HP1"]
    assert list(a.matches["HP1"]) == ["HP2", "HP3"]

    # test one to one
    a = ms.MSAligner(dataframe_1, dataframe_2, dataframe_3, names=names)
    a.align("HP1", "HP2")
    assert list(a.matches) == ["HP1"]
    assert list(a.matches["HP1"]) == ["HP2"]


def test_schema(dataframe_1, dataframe_2):
    dataframe_1 = dataframe_1.copy()
    dataframe_2 = dataframe_2.copy()
    schema_labels = {
        "Compound_ID": "feature_id",
        "RT": "rt (min)",
        "MZ": "m/z",
        "Intensity": "abundance",
    }

    a = ms.MSAligner(
        dataframe_1.rename(columns=schema_labels),
        dataframe_2.rename(columns=schema_labels),
        names=["DS1", "DS2"],
        schema_labels=schema_labels,
    )
    a.align()


def test_instance_params(dataframe_1, dataframe_2, dataframe_3):
    schema_labels = {"RT": "rt (min)"}
    dataframe_1 = dataframe_1.copy().rename(columns=schema_labels)
    dataframe_2 = dataframe_2.copy().rename(columns=schema_labels)
    dataframe_3 = dataframe_3.copy().rename(columns=schema_labels)

    a = ms.MSAligner(
        dataframe_1,
        dataframe_2,
        dataframe_3,
        names=["DS1", "DS2", "DS3"],
        schema_labels=schema_labels,
    )
    a.descriptors = {"rt (min)": "linear", "MZ": "ppm"}
    a.default_coarse_params["rt (min)"] = {"upper": 1.0, "lower": -1.0}
    a.default_cutoff = 10
    a.default_cutoffs = {"MZ": 6}
    a.align()

    a = ms.MSAligner(
        dataframe_1,
        dataframe_2,
        dataframe_3,
        names=["DS1", "DS2", "DS3"],
        schema_labels=schema_labels,
    )
    a.descriptors = {"rt (min)": "linear", "MZ": "ppm"}
    a.set_instance_defaults(
        {
            "coarse_params": {"rt (min)": {"upper": 1.0, "lower": -1.0}},
            "cutoff": 10,
            "cutoffs": {"MZ": 6},
        }
    )
    a.align()


@pytest.fixture()
def graph():
    """
    Digraph for testing compression
    """
    graph = nx.DiGraph()
    nodes = ["1__a", "2__a", "3__a", "4__a", "1__b", "2__b", "3__b", "4__b", "1__c"]
    graph.add_nodes_from(nodes)
    edges = [
        ("1__a", "2__a", {"score": 0.5, "rank": 0}),  # kept
        ("1__a", "3__a", {"score": 0.9, "rank": 0}),  # kept
        ("1__a", "4__a", {"score": 2, "rank": 0}),  # kept
        ("2__a", "1__a", {"score": 0.3, "rank": 0}),  # kept
        ("2__a", "3__a", {"score": 0.1, "rank": 0}),  # kept
        ("3__a", "1__a", {"score": 0.6, "rank": 0}),  # kept
        ("3__a", "2__a", {"score": 0.1, "rank": 0}),  # kept
        ("3__a", "4__a", {"score": 100, "rank": 0}),
        ("1__b", "2__a", {"score": 100, "rank": 0}),
        ("1__b", "3__b", {"score": 10, "rank": 0}),  # kept
        ("2__b", "1__b", {"score": 100, "rank": 0}),
        ("2__b", "3__a", {"score": 100, "rank": 0}),
        ("3__b", "1__b", {"score": 11, "rank": 0}),  # kept
        ("3__b", "2__b", {"score": 100, "rank": 0}),
        ("4__b", "2__b", {"score": 100, "rank": 0}),
        ("4__a", "1__a", {"score": 1, "rank": 0}),
        ("1__c", "4__a", {"score": 100, "rank": 0}),
    ]
    graph.add_edges_from(edges)
    return graph


def test_compress_graph(graph):
    """
    Test graph compression
    """

    compressed = ms.MSAligner.compress_graph(graph)

    # check nodes are present
    assert set(compressed.nodes) == set(graph.nodes)

    # check edges and weights
    edges_weights = set(
        frozenset([e[0], e[1], e[2]["score"]]) for e in compressed.edges(data=True)
    )
    assert edges_weights == {
        frozenset(["1__a", "2__a", 0.3 + 0.5]),
        frozenset(["2__a", "3__a", 0.1 + 0.1]),
        frozenset(["3__a", "1__a", 0.6 + 0.9]),
        frozenset(["4__a", "1__a", 1 + 2]),
        frozenset(["1__b", "3__b", 10 + 11]),
    }


@pytest.fixture()
def complex_compressed():
    """
    A complex feature graph
    """
    graph = nx.Graph()
    nodes = [
        ("1__a", {"dataset": 1}),
        ("2__a", {"dataset": 2}),
        ("3__a", {"dataset": 3}),
        ("4__a", {"dataset": 4}),
        ("5__a", {"dataset": 5}),
        ("1__b", {"dataset": 1}),
        ("2__b", {"dataset": 2}),
        ("3__b", {"dataset": 3}),
        ("4__b", {"dataset": 4}),
        ("1__c", {"dataset": 1}),
        ("2__c", {"dataset": 2}),
        ("3__d", {"dataset": 3}),
        ("4__c", {"dataset": 4}),
    ]
    graph.add_nodes_from(nodes)

    edges = [
        ("1__a", "2__a", {"score": 1}),
        ("3__a", "2__a", {"score": 1}),
        ("1__a", "3__a", {"score": 1}),
        ("4__a", "5__a", {"score": 0.7}),
        ("1__a", "5__a", {"score": 0.5}),
        ("1__a", "4__a", {"score": 0.3}),
        ("4__a", "2__b", {"score": 2}),
        ("4__a", "3__b", {"score": 2}),
        ("2__b", "3__b", {"score": 2}),
        ("4__b", "1__b", {"score": 0.5}),
        ("2__c", "1__b", {"score": 1}),
        ("3__d", "1__b", {"score": 1}),
        ("2__c", "3__d", {"score": 5}),
        ("3__b", "1__c", {"score": 1}),
        ("4__c", "1__c", {"score": 1}),
        ("5__b", "3__d", {"score": 0.1}),
        ("3__d", "4__b", {"score": 0.1}),
        ("2__c", "4__b", {"score": 0.1}),
        ("2__c", "5__b", {"score": 0.1}),
        ("1__c", "2__b", {"score": 0.1}),
    ]
    graph.add_edges_from(edges)
    return graph


@pytest.fixture()
def clique_compressed():
    """
    A simple clique graph
    """
    graph = nx.Graph()
    nodes = [
        ("1__a", {"dataset": 1}),
        ("2__a", {"dataset": 2}),
        ("3__a", {"dataset": 3}),
    ]
    graph.add_nodes_from(nodes)

    edges = [
        ("1__a", "2__a", {"score": 1}),
        ("3__a", "2__a", {"score": 1}),
        ("1__a", "3__a", {"score": 1}),
    ]
    graph.add_edges_from(edges)
    return graph


def test_deconvolution(complex_compressed, clique_compressed):
    """
    Test the deconvolution function
    """
    max_size = 3
    # test it returns a full clique
    result = ms.MSAligner.get_rank_groups(clique_compressed, max_size)
    assert len(result) == 1

    # test that it only reports 1 group
    result = ms.MSAligner.get_rank_groups(clique_compressed, max_size, 1, 1, 2)
    assert len(result) == 1

    max_size = 5
    # max cliques, 4, 3, >=1
    result = ms.MSAligner.get_rank_groups(complex_compressed, max_size)
    assert len(result) == 0
    result = ms.MSAligner.get_rank_groups(complex_compressed, max_size, 4, 4, 1)
    assert len(result) == 1
    result = ms.MSAligner.get_rank_groups(complex_compressed, max_size, 3, 3, 1)
    assert len(result) == 6
    result = ms.MSAligner.get_rank_groups(complex_compressed, max_size, 1, 1, 1)
    assert len(result) == 7

    # test the settings return the correct, sorted cliques
    expected = [
        ({"5__b", "2__c", "4__b", "3__d", "1__b"}, 5, 4, 8, 0.9874999999999998),
        ({"1__a", "3__a", "5__a", "2__a", "4__a"}, 5, 3, 6, 0.75),
        ({"1__a", "5__a", "3__b", "4__a", "2__b"}, 5, 3, 6, 1.25),
        ({"1__c", "2__b", "4__a", "3__b"}, 4, 3, 5, 1.42),
        ({"1__c", "4__c", "3__b", "2__b"}, 4, 3, 4, 1.025),
        ({"3__b", "1__b", "4__a", "2__b"}, 4, 3, 4, 1.75),
    ]
    results = ms.MSAligner.get_rank_groups(complex_compressed, max_size, 1, 1, 2)
    for i, r in enumerate(results):
        for x, y in zip(r, expected[i]):
            if isinstance(y, (float, int)):
                assert np.isclose(x, y)
            else:
                assert x == y


def test_integration(integration_data):
    """
    Tests alignment and 4 different aggregation methods
    """
    pickles = integration_data[0]
    datasets = integration_data[1]

    a = ms.MSAligner(*datasets, names=["DS1", "DS2", "DS3", "DS4"])
    a.align()

    results = a.results().loc[:, "DS1":"DS4"]
    results = results.sort_values(by=["DS1", "DS2", "DS3", "DS4"]).reset_index(
        drop=True
    )
    pickles[0].equals(results)

    results = a.results(c_size_or_loss=1, g_size_or_loss=1).loc[:, "DS1":"DS4"]
    results = results.sort_values(by=["DS1", "DS2", "DS3", "DS4"]).reset_index(
        drop=True
    )
    pickles[1].equals(results)

    results = a.results(c_size_or_loss=1, g_size_or_loss=1, remove_rerank=False).loc[
        :, "DS1":"DS4"
    ]
    results = results.sort_values(by=["DS1", "DS2", "DS3", "DS4"]).reset_index(
        drop=True
    )
    pickles[2].equals(results)

    results = a.results(c_size_or_loss=3, g_size_or_loss=3, diameter=2).loc[
        :, "DS1":"DS4"
    ]
    results = results.sort_values(by=["DS1", "DS2", "DS3", "DS4"]).reset_index(
        drop=True
    )
    pickles[3].equals(results)



# pylint: disable=redefined-outer-name, missing-function-docstring, consider-using-with
"""
Tests for blueshift
"""
import pickle
from pathlib import Path
import pytest
import pandas as pd
import numpy as np
from bmxp import blueshift as b


@pytest.fixture()
def path_dc_input_1():
    return Path(__file__).parent / "DCinput1.csv"


@pytest.fixture()
def path_sample_info_1():
    return Path(__file__).parent / "DCinfo1.csv"


@pytest.fixture()
def path_dc_input_2():
    return Path(__file__).parent / "DCinput2.csv"


@pytest.fixture()
def path_sample_info_2():
    return Path(__file__).parent / "DCinfo2.csv"


@pytest.fixture()
def df_dc_input_1(path_dc_input_1):
    return pd.read_csv(path_dc_input_1)


@pytest.fixture()
def df_sample_info_1(path_sample_info_1):
    return pd.read_csv(path_sample_info_1)


@pytest.fixture()
def df_dc_input_2(path_dc_input_2):
    return pd.read_csv(path_dc_input_2)


@pytest.fixture()
def df_sample_info_2(path_sample_info_2):
    return pd.read_csv(path_sample_info_2)


@pytest.fixture()
def pickled_results():
    return pd.read_pickle(Path(__file__).parent / "blueshift.pickle")


def test_data_validation(df_dc_input_1, df_sample_info_1):
    # missing required column in injection information
    info = df_sample_info_1.drop("injection_order", axis=1)
    with pytest.raises(ValueError) as e:
        b.DriftCorrection(df_dc_input_1, info)
    assert "injection_order" in str(e.value)

    # missing injection in data input
    data = df_dc_input_1.drop("B0005_COL_ExampleProject_CN-M36058078", axis=1)
    with pytest.raises(ValueError) as e:
        b.DriftCorrection(data, df_sample_info_1)
    assert "data sheet: B0005_COL_ExampleProject_CN-M36058078" in str(e.value)

    # no error when missing "not_used" injection in data input
    data = df_dc_input_1.drop("B0008_COL_ExampleProject_CN-M59244903", axis=1)
    b.DriftCorrection(data, df_sample_info_1)

    # duplicate injection order
    info = df_sample_info_1.copy()
    info.loc[14, "injection_order"] = info.loc[15, "injection_order"]
    with pytest.raises(ValueError) as e:
        b.DriftCorrection(df_dc_input_1, info)
    assert "duplicate values" in str(e.value)

    # duplicate injection id
    info = df_sample_info_1.copy()
    info.loc[14, "injection_id"] = info.loc[15, "injection_id"]
    with pytest.raises(ValueError) as e:
        b.DriftCorrection(df_dc_input_1, info)
    assert "duplicate injection_ids" in str(e.value)

    # out-of-order injection order
    info = df_sample_info_1.copy()
    info.loc[14, "injection_order"] = 700
    with pytest.raises(ValueError) as e:
        b.DriftCorrection(df_dc_input_1, info)
    assert "must be sorted" in str(e.value)

    # invalid label in batches column
    info = df_sample_info_1.copy()
    info.loc[13, "batches"] = "batch nd"
    with pytest.raises(ValueError) as e:
        b.DriftCorrection(df_dc_input_1, info)
    assert "invalid label" in str(e.value)

    # non-numeric character in data
    data = df_dc_input_1.copy()
    data.iloc[5, 5] = "f"
    with pytest.raises(TypeError) as e:
        b.DriftCorrection(data, df_sample_info_1)
    assert "non-numeric" in str(e.value)

    # data and sample are not in same order
    data = df_dc_input_1.copy()
    col_list = list(data.columns)
    col_list = col_list[:10] + col_list[11:] + col_list[10:11]
    data = data.loc[:, col_list]
    with pytest.raises(ValueError) as e:
        b.DriftCorrection(data, df_sample_info_1)
    assert "usable samples" in str(e.value)


def test_batch_start_end(df_sample_info_1):
    # batch_end shifts up to nearest valid injection
    info = df_sample_info_1.copy()
    info.loc[16, ["batches", "QCRole"]] = ["batch_end", "NA"]
    info[["batches", "QCRole"]] = info[["batches", "QCRole"]].fillna("")
    batches = b.find_batch_start_end(info)
    assert batches.loc[15] == "batch_end" and batches.loc[16] == ""

    info = df_sample_info_1.copy()
    info.loc[7, "batches"] = "batch_end"
    info.loc[:7, "QCRole"] = "NA"
    info[["batches", "QCRole"]] = info[["batches", "QCRole"]].fillna("")
    with pytest.raises(ValueError) as e:
        b.find_batch_start_end(info)
    assert "Cannot move " in str(e.value)


def test_batch_generation(
    df_dc_input_1,
    df_sample_info_1,
    path_dc_input_2,
    path_sample_info_2,
    pickled_results,
):
    a = b.DriftCorrection(df_dc_input_1, df_sample_info_1)
    for batch, ref_batch in zip(a.batches["default"], pickled_results["default1"]):
        assert (batch.values == ref_batch.values).all()
    for batch, ref_batch in zip(a.batches["override"], pickled_results["override1"]):
        assert (batch.values == ref_batch.values).all()

    a = b.DriftCorrection(path_dc_input_2, path_sample_info_2)
    for batch, ref_batch in zip(a.batches["default"], pickled_results["default2"]):
        assert (batch.values == ref_batch.values).all()
    for batch, ref_batch in zip(a.batches["override"], pickled_results["override2"]):
        assert (batch.values == ref_batch.values).all()


def test_internal_standard_correction(
    path_dc_input_1,
    df_dc_input_1,
    path_sample_info_1,
    df_dc_input_2,
    df_sample_info_2,
    pickled_results,
):
    # one internal standard
    a = b.DriftCorrection(path_dc_input_1, path_sample_info_1)
    a.internal_standard_correct("Internal Standard 1")
    assert np.isclose(
        a.data.round().fillna(0),
        pickled_results["DCinput1_IS_InternalStandard1"].round().loc[:, a.data.columns],
        equal_nan=True,
    ).all()

    # one internal standard with nonquant duplicate
    nonquant_df = df_dc_input_1.copy()
    nonquant_df.loc[4, "Metabolite"] = "Internal Standard 1"
    nonquant_df.loc[4, "Non_Quant"] = True
    a = b.DriftCorrection(nonquant_df, path_sample_info_1)
    a.internal_standard_correct("Internal Standard 1")
    assert np.isclose(
        a.data.round().fillna(0),
        pickled_results["DCinput1_IS_InternalStandard1"].round().loc[:, a.data.columns],
        equal_nan=True,
    ).all()

    # nonquant "missing" internal standard
    nonquant_df = df_dc_input_1.copy()
    nonquant_df.loc[0, "Non_Quant"] = True
    a = b.DriftCorrection(nonquant_df, path_sample_info_1)
    with pytest.raises(ValueError) as e:
        a.internal_standard_correct("Internal Standard 1")
    assert "not found in" in str(e.value)

    # two internal standards
    a = b.DriftCorrection(df_dc_input_2, df_sample_info_2)
    a.internal_standard_correct(["15R-15-methyl-PGA2", "15R-15-methyl-PGF2a"])
    assert np.isclose(
        a.data.round(),
        pickled_results["DCinput2_IS_PGA2_PGF2a"].loc[:, a.data.columns],
        equal_nan=True,
    ).all()

    # missing IS value
    data = df_dc_input_2.copy()
    data.iloc[14, 50] = 0
    a = b.DriftCorrection(data, df_sample_info_2)
    with pytest.raises(ValueError) as e:
        a.internal_standard_correct("15S-15-methyl-PGD2")
    assert "missing values" in str(e.value)

    # wrong IS name
    with pytest.raises(ValueError) as e:
        a.internal_standard_correct("not_a_real_metabolite")
    assert "not found in" in str(e.value)


def test_pool_correction(
    path_dc_input_1,
    path_sample_info_1,
    path_dc_input_2,
    path_sample_info_2,
    pickled_results,
):
    # linear with override
    a = b.DriftCorrection(path_dc_input_1, path_sample_info_1)
    a.pool_correct(
        interpolation="linear", pool="PREFA", override=True, max_missing_percent=100
    )
    assert np.isclose(
        a.data.apply(np.floor),
        pickled_results["DCinput1_linear_PREFA_override"].loc[:, a.data.columns],
        equal_nan=True,
    ).all()

    # linear without override
    a = b.DriftCorrection(path_dc_input_1, path_sample_info_1)
    a.pool_correct(
        interpolation="linear", pool="PREFA", override=False, max_missing_percent=100
    )
    assert np.isclose(
        a.data.apply(np.floor),
        pickled_results["DCinput1_linear_PREFA"].loc[:, a.data.columns],
        equal_nan=True,
    ).all()

    # internal standard + NN
    a = b.DriftCorrection(path_dc_input_2, path_sample_info_2)
    a.internal_standard_correct("15R-15-methyl-PGA2")
    a.pool_correct(interpolation="NN", pool="PREFB", max_missing_percent=100)
    assert np.isclose(
        a.data.round(),
        pickled_results["DCinput2_IS_PGA2_NN_PREFB"].loc[:, a.data.columns],
        equal_nan=True,
    ).all()

    # linear with max_missing_percent=30
    a = b.DriftCorrection(path_dc_input_1, path_sample_info_1)
    a.pool_correct(
        interpolation="linear", pool="PREFA", override=True, max_missing_percent=30
    )
    assert np.isclose(
        a.data.apply(np.floor),
        pickled_results["DCinput1_linear_PREFA_override_maxmissing30"]
        .loc[:, a.data.columns]
        .apply(np.floor),
        equal_nan=True,
    ).all()


def test_cv_calculation(
    path_dc_input_2,
    path_sample_info_2,
    pickled_results,
):
    # CV calculation
    a = b.DriftCorrection(path_dc_input_2, path_sample_info_2)
    a.pool_correct(interpolation="linear", pool="PREFA", max_missing_percent=100)
    a.calculate_cvs()
    res = a.cvs.loc[:, ["CV" in col for col in a.cvs.columns]]
    assert np.isclose(
        res.fillna(0),
        pickled_results["DCinput2_linear_PREFA_CVs"].loc[:, res.columns].fillna(0),
    ).all()



"""Definitions of dataclasses used by the python rawfilereader"""

# pylint: disable=too-many-instance-attributes,missing-class-docstring
from dataclasses import dataclass, field
from typing import List
from rawfilereader import Analyzer, ScanMode, ScanType, Polarity


@dataclass
class Xic:
    rt: float
    intensity: float
    length: int
    mz: float


@dataclass
class ProfileChunk:
    first_bin: int = 0
    n_bins: int = 0
    fudge: float = 0
    signal: List[float] = field(default_factory=list)


@dataclass
class Scan:
    intensities: List[float] = field(default_factory=list)
    mzs: List[float] = field(default_factory=list)
    peak_total: int = 0  # granular total of all mz/ints

    cent_intensities: List[float] = field(default_factory=list)
    cent_mzs: List[float] = field(default_factory=list)

    chunks: List[ProfileChunk] = field(default_factory=list)

    first_value: float = 0
    step: float = 0
    peak_count: int = 0  # number of chunks (peaks?)
    offset: int = 0
    a: float = 0
    b: float = 0
    c: float = 0  # pylint: disable=invalid-name
    count: int = 0  # centroided peak count
    peak_list_size: int = 0
    layout: int = 0
    n_param: int = 0
    n_bins: int = 0
    profile_size: int = 0
    time: float = 0
    index: int = 0
    filter: int = -1  # index of matching filter in RawFile.scan_filters


@dataclass
class ScanFilter:
    polarity: Polarity = Polarity.UNDEFINED
    scan_mode: ScanMode = ScanMode.UNDEFINED
    dependent_scans: bool = None
    ms_level: int = None
    scan_type: ScanType = ScanType.UNDEFINED
    analyzer: Analyzer = Analyzer.UNDEFINED
    precursor_mz: List[float] = field(default_factory=list)
    energy: List[float] = field(default_factory=list)
    low_mass: float = None
    high_mass: float = None
    filter_string: str = ""


@dataclass
class RawFile:
    data: bytes = b""
    instrument_model: str = ""
    file_name: str = ""
    file_format: str = "raw"
    scans: List[Scan] = field(default_factory=list)
    num_scans: int = 0
    scan_trailer_addr: int = 0
    first_scan_number: int = 0
    last_scan_number: int = 0
    low_mz: float = 0
    high_mz: float = 0
    start_time: float = 0
    end_time: float = 0
    scan_index_addr: int = 0
    scan_params_addr: int = 0
    version: int = 0
    n_controllers: int = 0
    data_addr: int = 0
    scan_filters: List[ScanFilter] = field(default_factory=list)
    timestamp: int = 0



# pylint: skip-file
"""A python implementation of the C finniganreader"""

import struct
from rawfilereader import Analyzer, ScanMode, ScanType, Polarity
from rawfilereader.main_dataclasses import Xic, ProfileChunk, Scan, ScanFilter, RawFile
from rawfilereader.mzml_parser import parse_mzml


def from_pascal(data, pos):
    """
    Decode a pascal string from bytes
    :param data: bytes
    :param pos: index of the beginning of the string
    :return: string
    """
    [length] = struct.unpack("I", data[pos : pos + 4])
    pos += 4
    pascal = []
    for _ in range(length):
        pascal += data[pos : pos + 1]
        pos += 2
    return bytes(pascal).decode("utf-8")


def move_pascal(data, pos, n):
    """
    Skips past n pascal strings, beginning at index=pos
    :param data: bytes
    :param pos: index of the beginning of the first string
    :param n: number of strings to skip over
    :return: position after skipping over n strings
    """
    for _ in range(n):
        [length] = struct.unpack("I", data[pos : pos + 4])
        pos += 4 + length * 2
    return pos


def fill_scan_events(rawfile):
    """
    Create Scans and populate with basic scan event info
    :param rawfile:
    """
    pos = rawfile.scan_trailer_addr + 4

    for _ in range(rawfile.num_scans):
        scan_i = Scan()
        scan_filter = ScanFilter()
        # scan_mode at pos+5, ms_level at pos+6, scan_type at pos+7, data_dependent at
        # pos+10, analyzer at pos+40
        [
            polarity,
            scan_mode,
            scan_filter.ms_level,
            scan_type,
            dependent_scans,
            analyzer,
        ] = struct.unpack("4B2xB29xB", rawfile.data[pos + 4 : pos + 41])

        scan_filter.polarity = Polarity(polarity)
        scan_filter.scan_mode = ScanMode(scan_mode)
        scan_filter.scan_type = ScanType(scan_type)
        scan_filter.analyzer = Analyzer(analyzer)
        scan_filter.dependent_scans = bool(dependent_scans)

        pos += 136
        [n_precursors] = struct.unpack("I", rawfile.data[pos : pos + 4])
        pos += 4

        # in non-ID-X data, MS1 data has n_precursors=1 and 56 bytes of unknowns to skip
        # in ID-X data, MS1 data has n_precursors=0 and 56 bytes of data/unknowns
        for _ in range(n_precursors):
            if scan_filter.ms_level > 1:
                scan_filter.precursor_mz.append(
                    round(struct.unpack("d", rawfile.data[pos : pos + 8])[0], 4)
                )
                scan_filter.energy.append(
                    struct.unpack("d", rawfile.data[pos + 16 : pos + 24])[0]
                )
            pos += 56

        [scan_filter.low_mass, scan_filter.high_mass, scan_i.n_param] = struct.unpack(
            "ddI", rawfile.data[pos + 4 : pos + 24]
        )
        pos += 24

        if scan_i.n_param > 0:
            [scan_i.a, scan_i.b, scan_i.c] = struct.unpack(
                "3d", rawfile.data[pos + 16 : pos + 40]
            )

        if rawfile.instrument_model == "Orbitrap Exploris 240":
            pos += 48
            pos = move_pascal(rawfile.data, pos, 1)
        elif scan_i.n_param == 7:  # Orbitrap ID-X
            pos += 68
        elif scan_i.n_param == 5:
            pos += 60
        elif scan_i.n_param == 0:
            pos += 12

        for j, filter_j in enumerate(rawfile.scan_filters):
            if filter_j == scan_filter:
                scan_i.filter = j
                break
        if scan_i.filter == -1:
            scan_i.filter = len(rawfile.scan_filters)
            rawfile.scan_filters.append(scan_filter)

        scan_i.filter = rawfile.scan_filters.index(scan_filter)

        rawfile.scans.append(scan_i)


def fill_scan_indices(rawfile):
    """
    Populate scans with index, time, offset
    :param rawfile:
    """
    pos = rawfile.scan_index_addr
    for i in range(rawfile.num_scans):
        # index at pos+4, time at pos+24, offset at pos+72
        [
            rawfile.scans[i].index,
            rawfile.scans[i].time,
            rawfile.scans[i].offset,
        ] = struct.unpack("=I16xd40xQ", rawfile.data[pos + 4 : pos + 80])
        pos += 88


def read_scan_data_packet(rawfile, index, scan, profile):
    """
    Populate scan with actual peaks/data
    :param rawfile:
    :param index: address/index where scan data begins (scan.offset)
    :param scan: scan to be read
    """
    [scan.profile_size, scan.peak_list_size, scan.layout] = struct.unpack(
        "3I", rawfile.data[index + 4 : index + 16]
    )
    index += 40
    total_peaks = 0
    if scan.profile_size > 0:
        [scan.first_value, scan.step, scan.peak_count, scan.n_bins] = struct.unpack(
            "ddII", rawfile.data[index : index + 24]
        )
        index += 24
        if not profile:
            for _ in range(scan.peak_count):
                chunk = ProfileChunk()
                [chunk.first_bin, chunk.n_bins] = struct.unpack(
                    "II", rawfile.data[index : index + 8]
                )
                index += 8
                if scan.layout > 0:
                    index += 4
                index += 4 * chunk.n_bins
        else:
            for _ in range(scan.peak_count):
                chunk = ProfileChunk()
                [chunk.first_bin, chunk.n_bins] = struct.unpack(
                    "II", rawfile.data[index : index + 8]
                )
                total_peaks += chunk.n_bins
                index += 8
                if scan.layout > 0:
                    [chunk.fudge] = struct.unpack("f", rawfile.data[index : index + 4])
                    index += 4
                chunk.signal = list(
                    struct.unpack(
                        f"{chunk.n_bins}f",
                        rawfile.data[index : index + 4 * chunk.n_bins],
                    )
                )
                index += 4 * chunk.n_bins
                scan.chunks.append(chunk)

    if scan.peak_list_size > 0:
        [scan.count] = struct.unpack("I", rawfile.data[index : index + 4])
        index += 4
        for _ in range(scan.count):
            [mz, intensity] = struct.unpack("ff", rawfile.data[index : index + 8])
            scan.cent_intensities.append(intensity)
            scan.cent_mzs.append(mz)
            index += 8

    scan.peak_total = total_peaks


def fill_scans(rawfile, profile, centroid):
    """
    Populate scans with usable MZs/intensities
    :param rawfile:
    """
    for i in range(rawfile.num_scans):
        scan = rawfile.scans[i]
        read_scan_data_packet(rawfile, scan.offset, scan, profile)
        if not profile:
            scan.intensities = []
            scan.mzs = []
            continue
        # convert profile readings into mz/intensities
        for peak_n in range(scan.peak_count):
            for bin_n in range(scan.chunks[peak_n].n_bins):
                # convert Hz values to m/z
                v = (
                    scan.first_value
                    + (scan.chunks[peak_n].first_bin + bin_n) * scan.step
                )
                if scan.n_param == 4:
                    scan.mzs.append(
                        scan.a + scan.b / v + scan.c / v / v + scan.chunks[peak_n].fudge
                    )
                elif scan.n_param in (5, 7):
                    scan.mzs.append(
                        scan.a
                        + scan.b / v / v
                        + scan.c / v / v / v / v
                        + scan.chunks[peak_n].fudge
                    )
                else:
                    scan.mzs.append(v + scan.chunks[peak_n].fudge)
            scan.intensities.extend(scan.chunks[peak_n].signal)
        scan.chunks = []

        # # put centroided data in mz/intensity lists if there's no profile data
        # if scan.peak_count == 0:
        #     scan.mzs = scan.cent_mzs.copy()
        #     scan.intensities = scan.cent_intensities.copy()


def initialize(rawfile, profile, centroid):
    """
    Fetch all relevant info from the provided rawfile
    :param rawfile:
    """
    # Skip to the version number in FileHeader
    pos = 36
    [rawfile.version] = struct.unpack("I", rawfile.data[pos : pos + 4])
    [rawfile.timestamp] = struct.unpack("I", rawfile.data[pos + 4 : pos + 12])
    print(f"version: {rawfile.version}")
    if rawfile.version < 66:
        raise ValueError("Versions under 66 are not supported.")
    pos = 1420  # start right after InjectionData in SequencerRow
    pos = move_pascal(rawfile.data, pos, 11)  # skip labels, comments, method files
    rawfile.file_name = from_pascal(rawfile.data, pos)
    pos = move_pascal(rawfile.data, pos, 5)  # skip filename, path, vial name
    pos += 4
    pos = move_pascal(rawfile.data, pos, 15)
    pos += 24  # AutoSamplerInfo
    pos = move_pascal(rawfile.data, pos, 1)
    pos += 28  # skip the methodfile and timestamp
    [rawfile.n_controllers] = struct.unpack("I", rawfile.data[pos : pos + 4])
    pos += 780  # skip other stuff in the preamble and the padding
    [rawfile.data_addr] = struct.unpack("Q", rawfile.data[pos : pos + 8])
    pos += 16
    run_header_addr = []
    for i in range(rawfile.n_controllers):
        run_header_addr.append(struct.unpack("Q", rawfile.data[pos : pos + 8])[0])
        pos += 16  # skip runHeaderAddr and the unknown 7
    rawfile.scan_trailer_addr = 0
    for i in range(rawfile.n_controllers):
        if rawfile.scan_trailer_addr != 0:
            break
        pos = run_header_addr[i]
        [rawfile.first_scan_number, rawfile.last_scan_number] = struct.unpack(
            "II", rawfile.data[pos + 8 : pos + 16]
        )
        [
            rawfile.low_mz,
            rawfile.high_mz,
            rawfile.start_time,
            rawfile.end_time,
        ] = struct.unpack("4d", rawfile.data[pos + 56 : pos + 88])
        [rawfile.scan_index_addr] = struct.unpack(
            "Q", rawfile.data[pos + 7408 : pos + 7416]
        )
        [rawfile.scan_trailer_addr, rawfile.scan_params_addr] = struct.unpack(
            "QQ", rawfile.data[pos + 7448 : pos + 7464]
        )
        pos += 7588
        pos = move_pascal(rawfile.data, pos, 1)
        rawfile.instrument_model = from_pascal(rawfile.data, pos)

    # pull the scan Events
    rawfile.num_scans = rawfile.last_scan_number - rawfile.first_scan_number + 1

    if rawfile.n_controllers == 0 or rawfile.num_scans == 0:
        raise ValueError("Invalid raw file.")

    fill_scan_events(rawfile)
    fill_scan_indices(rawfile)
    for j in range(rawfile.num_scans):
        rawfile.scans[j].offset += rawfile.data_addr
    fill_scans(rawfile, profile, centroid)


def find_first_ge(arr, target_rt):
    """
    Find the first RT that is greater than or equal to the target RT
    :param arr: list of scans
    :param target_rt: RT to search for
    :return: int, index of resulting RT
    """
    n = len(arr)
    if arr[0].time >= target_rt:
        return 0
    if n == 1:
        return 1
    midpoint = int(n / 2)
    if arr[midpoint].time >= target_rt:
        return find_first_ge(arr[:midpoint], target_rt)
    return midpoint + find_first_ge(arr[midpoint:], target_rt)


def find_first_ge_float(arr, target):
    """
    find_first_ge, but generalized to find any target in any list
    :param arr: list of values to search
    :param target: target value to find
    :return: int, index of resulting value
    """
    n = len(arr)
    if n == 0 or arr[0] >= target:
        return 0
    if n == 1:
        return 1
    midpoint = int(n / 2)
    if arr[midpoint] >= target:
        return find_first_ge_float(arr[:midpoint], target)
    return midpoint + find_first_ge_float(arr[midpoint:], target)


def pull_xic(
    rawfile, rt1, rt2, mz, ppm=20, scan_filter=None, centroid=True, pull_mzs=False
):
    """
    Generate RTs/intensities for a given MZ and given RT range
    :param rawfile:
    :param rt1: float, lower end of RT range
    :param rt2: float, upper end of RT range
    :param mz: float, target MZ
    :param ppm: float, allowed delta ppm from target MZ
    :param scan_filter: int, index of filter in RawFile.scan_filters list
    :param centroid: bool, use centroided peaks instead of profile data
    :return: Xic containing extracted data
    """
    start = find_first_ge(rawfile.scans, rt1)
    stop = find_first_ge(rawfile.scans, rt2) - 1
    length = stop - start + 1
    intensity = []
    rt = []
    return_mzs = []

    for i in range(length):
        if scan_filter is None or rawfile.scans[start + i].filter == scan_filter:
            min_mz = mz - (mz / 1000000) * ppm
            max_mz = mz + (mz / 1000000) * ppm

            rt.append(rawfile.scans[start + i].time)
            intensity.append(0)
            return_mzs.append(0)

            if centroid:
                mzs = rawfile.scans[start + i].cent_mzs
                intensities = rawfile.scans[start + i].cent_intensities
            else:
                mzs = rawfile.scans[start + i].mzs
                intensities = rawfile.scans[start + i].intensities

            peak_start = find_first_ge_float(mzs, min_mz)
            peak_stop = find_first_ge_float(mzs, max_mz) - 1

            if peak_start > peak_stop:
                continue
            avg_mz = 0
            for j in range(peak_stop - peak_start + 1):
                intensity[-1] += intensities[j + peak_start]
                if pull_mzs:
                    avg_mz += intensities[j + peak_start] * mzs[j + peak_start]
            if pull_mzs:
                try:
                    return_mzs[-1] = avg_mz / intensity[-1]
                except ZeroDivisionError:
                    return_mzs[-1] = 0

    return Xic(rt=rt, intensity=intensity, length=length, mz=return_mzs)


def from_bytes(data, profile=False, centroid=True):
    """
    Generate a rawfile from bytes
    :param data: bytes, rawfile contents
    :return: rawfile
    """
    rawfile = RawFile(data=data)
    initialize(rawfile, profile, centroid)
    rawfile.data = b""
    return rawfile


def open_rawfile(filename, profile=False, centroid=True):
    """
    Open a rawfile from a file name
    :param filename: string, filename
    :return: rawfile
    """
    with open(filename, mode="rb") as file:
        data = file.read()
    rawfile = from_bytes(data, profile, centroid)
    return rawfile


def open_mzml(filename_or_stream):
    """
    Open an mzML file from a file name. Wraps parse_mzml so it can be accessed from the
    main module.
    :param filename_or_stream: string (filename) or file object
    :return: rawfile
    """
    return parse_mzml(filename_or_stream)


if __name__ == "__main__":
    # for i in range(1):
    #     print(i)
    #     rawfile = open_rawfile(r"D:\Work\TFE Demo\PREFA05.raw")
    # exit()
    # rawfile = open_rawfile(r"D:\Work\TFE Demo\PREFA05.raw")

    rawfile = open_rawfile(r"D:\Work\TFE Demo\PREFA05.raw", profile=True, centroid=True)
    result = pull_xic(rawfile, 9.76, 9.8, 89.1073)
    print(result.intensity)
    print(result.mz)
    result = pull_xic(rawfile, 9.76, 9.8, 89.1073, pull_mzs=True)
    print(result.intensity)
    print(result.mz)
    result = pull_xic(rawfile, 9.76, 9.8, 89.1073, centroid=False)
    print(result.intensity)
    print(result.mz)
    result = pull_xic(rawfile, 9.76, 9.8, 89.1073, centroid=False, pull_mzs=True)
    print(result.intensity)
    print(result.mz)
